/***************************************************************************

  ratecnv.c

  Sampling rate converter

***************************************************************************/

#include "driver.h"
#include <stdio.h>
#include <math.h>
#include "ratecnv.h"


#define X86_USE_ASM_MMX
#define ERRORLOG


#define CACHE_LINE_SIZE    (1 << 6)     // 64 bytes, for align


#define LPF_MAX_CHANNELS    32
#define SRCNV_MAX_CHANNELS  32


static int use_filter = 1;

/********************************************************************
    lpf param
*********************************************************************/

#define _A_             102             // 102db
#define LPF_TAPS        (1 << 8)        // 256 TAPS
#define LPF_TAPS_MASK   (LPF_TAPS - 1)  // 0xFF
#define LPF_FRAC        14              // Q14 format

typedef struct{
    INT16 *ringbuff;
    INT32 ringbuff_top;
    INT16 *h;
    INT32 taps;
    INT32 tapsmask;
    INT32 lpf_fracbits;
}LPF_SUB;

static int srcnv_init_flag = 0;

static double lpf_temp[LPF_TAPS + 1];
static INT16 lpf_p0[LPF_MAX_CHANNELS];

static INT16 *p_lpf_p;
static INT8  *p_lpf_p_ = 0;

static int lpf_top[LPF_MAX_CHANNELS];
static int lpf_current_used = 0;



/********************************************************************
    srcnv param
*********************************************************************/

#define SRCNV_FC_ADJUST_RATIO     0.08

#define LINEAR_IP_FRACBITS          12
#define LINEAR_IP_FRACBITS_MASK     ((1 << LINEAR_IP_FRACBITS) - 1)

#define UPSAMPLING_FRAC         30                              // Q30 format for delta
#define UPSAMPLING_FRAC_MASK    ((1 << UPSAMPLING_FRAC) - 1)

#define SRCNV_BUFFER_LENGTH     2048
static INT16 srcnv_buffer[SRCNV_BUFFER_LENGTH];
static int srcnv_current_used = 0;

typedef struct{
    int no_op;
    int lpf_channel;
    int down_step;
    int down_step_counter;
    double src_freq;
    double dst_freq;
    UINT64 src_dst_ratio;           // Q30 format

/* for upsampling */
    UINT32 upsampling_current;      // Q30 format
    UINT32 upsampling_delta;        // Q30 format
    int data_top;
    INT16 data[2];

    INT16 *src;
    INT16 *srcend;
    INT16 *temp;
    int    tempbufferlen;
    int    upsampling_outlen;
    INT16 *outbuffer;
    int    outbufferlen;

    int finished;
} SRCNV;

static SRCNV srcnv_data[SRCNV_MAX_CHANNELS];


/********************************************************************
                            LPF section
********************************************************************/

static INT16* lpf_get_p_pointer(int lpf_channel, int number)
{
    INT16* a0;
    int d0;

    if(lpf_channel < 0 || lpf_current_used <= lpf_channel) return 0;

    number = number & 0x03;
    d0 = ((lpf_channel * 5) + number) * LPF_TAPS;
    a0 = p_lpf_p + d0;

    return a0;
}


static INT16* lpf_get_ring_buffer(int lpf_channel)
{
    INT16* a0;
    int d0;

    if(lpf_channel < 0 || lpf_current_used <= lpf_channel) return 0;

    d0 = ((lpf_channel * 5) + 4) * LPF_TAPS;
    a0 = p_lpf_p + d0;

    return a0;
}


static int lpf_init_buffer(void)
{
    int bytes;
    INT8 *temp;

    bytes  = LPF_MAX_CHANNELS * 2 * LPF_TAPS * (4 + 1);  // (filter param * 4) + ring buff
    bytes += CACHE_LINE_SIZE;                            // for align CACHE_LINE_SIZE
    p_lpf_p_ = (INT8*)malloc(bytes);
    if(p_lpf_p_ == 0) return 0;
    temp = p_lpf_p_ + (CACHE_LINE_SIZE - 1);
    temp = (INT8*)((unsigned long)temp & ~(CACHE_LINE_SIZE - 1));
    p_lpf_p = (INT16*)temp;

    return 1;
}


static void lpf_exit_buffer(void)
{
    if(p_lpf_p_ != 0){
        free(p_lpf_p_);
        p_lpf_p_ = 0;
    }
}



static double bf0(double x)
{
    double sum;
    double f0;
    int d0;

    sum = 1.0;
    f0  = 1.0;
    d0 = 1;
    while(d0 < 20){
        f0  = (f0 * 0.5 * x) / d0;
        sum = sum + (f0 * f0);
        d0 += 1;
    }
    return (sum);
}


static void fir_init(double fs, double fc, double p[])
{
    double sum;
    double a;
    double w;
    double f;
    int d0;
    int center;

    a = 0.1102*(_A_ - 8.7);
    p[0] = (fc / fs) * 2.0;
    sum = p[0];
    d0 = 1;
    while(d0 <= LPF_TAPS/2){
        f = (d0 * 2.0) / LPF_TAPS;
        w = bf0(a * sqrt(1.0 - (f * f))) / bf0(a);
        p[d0] = (1.0 / (M_PI*d0)) * sin(d0 * 2.0 * M_PI * (fc / fs)) * w;
        sum += p[d0] * 2.0;
        d0 += 1;
    }
    d0 = 0;
    while(d0 <= LPF_TAPS/2){
        p[d0] = p[d0]/sum;
        d0 += 1;
    }

    /* copy */
    center = LPF_TAPS/2;
    d0 = LPF_TAPS/2;
    while(d0 >= 0){
        p[center + d0] = p[d0];
        d0 -= 1;
    }
    d0 = 1;
    while(d0 <= LPF_TAPS/2){
        p[center - d0] = p[center + d0];
        d0 += 1;
    }
}


static int lpf_init_sub(int lpf_channel, double fs, double fc, int init_flag)
{
    int d0;
    int d1;
    INT16 *a0;

    if(lpf_channel < 0 || lpf_current_used <= lpf_channel) return -1;
    if(use_filter == 0) return lpf_channel;
    if(fs <= 0){
#ifdef ERRORLOG
        logerror("LPF ch %d, fs <= 0\n", lpf_channel);
#endif
        return -1;
    }
    if(fc <= 0){
#ifdef ERRORLOG
        logerror("LPF ch %d, fc <= 0\n", lpf_channel);
#endif
        return -1;
    }
//    if(init_flag != 0){
    if(1){
        lpf_top[lpf_channel] = 0;
        a0 = lpf_get_ring_buffer(lpf_channel);
        d0 = 0;
        while(d0 < LPF_TAPS){
            a0[d0] = 0;
            d0 += 1;
        }
    }

    if(fc <= 1) fc = 1;
#ifdef ERRORLOG
    logerror("LPF ch:%d  fs:%d  fc:%d\n", lpf_channel, (int)fs, (int)fc);
#endif
    fir_init(fs, fc, &lpf_temp[0]);

    d0 = 0;
    while(d0 <= LPF_TAPS){
        lpf_temp[d0] = lpf_temp[d0] * (1 << LPF_FRAC);          // Q14 format
        d0 += 1;
    }

    lpf_p0[lpf_channel] = lpf_temp[0];      // Q14 format

    /* buffer 0 */
    a0 = lpf_get_p_pointer(lpf_channel, 0);
    d1 = 0;
    d0 = 1;
    while(d0 <= LPF_TAPS){
        a0[d1++] = lpf_temp[d0];            // Q14 format
        d0 += 1;
    }

    /* buffer 3 */
    a0 = lpf_get_p_pointer(lpf_channel, 3);
    d1 = 0;
    a0[d1++] = lpf_temp[LPF_TAPS - 2];
    a0[d1++] = lpf_temp[LPF_TAPS - 1];
    a0[d1++] = lpf_temp[LPF_TAPS - 0];
    a0[d1++] = lpf_temp[1];
    d0 = 2;
    while(d0 <= (LPF_TAPS - 3)){
        a0[d1++] = lpf_temp[d0];
        d0 += 1;
    }

    /* buffer 2 */
    a0 = lpf_get_p_pointer(lpf_channel, 2);
    d1 = 0;
    a0[d1++] = lpf_temp[LPF_TAPS - 1];
    a0[d1++] = lpf_temp[LPF_TAPS - 0];
    a0[d1++] = lpf_temp[1];
    a0[d1++] = lpf_temp[2];
    d0 = 3;
    while(d0 <= (LPF_TAPS - 2)){
        a0[d1++] = lpf_temp[d0];            // Q14 format
        d0 += 1;
    }

    /* buffer 1 */
    a0 = lpf_get_p_pointer(lpf_channel, 1);
    d1 = 0;
    a0[d1++] = lpf_temp[LPF_TAPS - 0];
    a0[d1++] = lpf_temp[1];
    a0[d1++] = lpf_temp[2];
    a0[d1++] = lpf_temp[3];
    d0 = 4;
    while(d0 <= (LPF_TAPS - 1)){
        a0[d1++] = lpf_temp[d0];            // Q14 format
        d0 += 1;
    }

    return lpf_channel;
}


int lpf_new(double fs, double fc)
{
    int lpf_channel;

    if(srcnv_init_flag == 0) srcnv_init(1);         /* force to use lpf */

    if(lpf_current_used >= LPF_MAX_CHANNELS) return -1;
    lpf_channel = lpf_current_used;
    lpf_current_used += 1;
#ifdef ERRORLOG
    logerror("LPF_NEW ch:%d, used:%d\n", lpf_channel, lpf_current_used);
#endif
    lpf_init_sub(lpf_channel, fs, fc, 1);

    return lpf_channel;
}


int lpf_change(int lpf_channel, double fs, double fc)
{
#ifdef ERRORLOG
    logerror("LPF_CHANGE: ch:%d\n", lpf_channel, lpf_current_used);
#endif
    lpf_init_sub(lpf_channel, fs, fc, 0);

    return lpf_channel;
}



#ifdef X86_USE_ASM_MMX

extern INT16 lpf_apply_calc_asm(
            LPF_SUB *ls,
            INT32 p0s
);

extern void do_emms(void);

#else

static INT16 lpf_apply_calc(LPF_SUB *ls, INT32 p0s)
{
/*********************************************************************
goal:
    sum = 0;
    i = 0;
    while(i < TAPS){
        sum += x[i] * h[i];
        i += 1;
    }
    return sum;

means:
    for SIMD, x[i] can be the following sequences.

0       x[0]        x[TAPS-1]   x[TAPS-2]   x[TAPS-3]
1       x[1]        x[0]        x[TAPS-1]   x[TAPS-2]
2       x[2]        x[1]        x[0]        x[TAPS-1]
3       x[3]        x[2]        x[1]        x[0]

4       x[4]        x[3]        x[2]        x[1]
5       x[5]        x[4]        x[3]        x[2]
6       x[6]        x[5]        x[4]        x[3]
7       x[7]        x[6]        x[5]        x[4]

8       x[8]        x[7]        x[6]        x[5]
.       .           .           .           .
.       .           .           .           .
.       .           .           .           .
TAPS-1  x[TAPS-1]   x[TAPS-2]   x[TAPS-3]   x[TAPS-4]

x[i] and h[i] must be the same sequence.
TAPS must be a power of 2.
TAPS must be equal to 16, or greater than 16.
*********************************************************************/

    int d0;
    int d1;
    INT32 sum;
    INT16 *x;
    INT16 *h;

    x  = ls->ringbuff;
    h  = ls->h;
    d1 = ls->ringbuff_top;
    sum = p0s;
    d0 = 0;
    while(d0 < ls->taps){
        sum += x[d1 + 0] * h[d0 + 0];
        sum += x[d1 + 1] * h[d0 + 1];
        sum += x[d1 + 2] * h[d0 + 2];
        sum += x[d1 + 3] * h[d0 + 3];
        d1 = (d1 + 4) & ls->tapsmask;
        d0 += 4;
    }
    sum = sum >> ls->lpf_fracbits;      // Q14 format to INT16
    if(sum >  32767) sum =  32767;
    if(sum < -32768) sum = -32768;

    return (INT16)sum;
}


#endif



static int lpf_apply_sub(int lpf_channel, INT16 *src, int length,
                                    int *downstep, int *downstepcounter)
{
//      return value is outlength

    LPF_SUB ls;
    INT16 *ringbuff;
    INT16 *dst;
    int dstlen;
    int top_pos;
    int d0;
    INT16 sample;
    INT16 p0;
    int dstep;
    int dstepcounter;

    if(lpf_channel < 0) return 0;

    dstep        = *downstep;
    dstepcounter = *downstepcounter;
    if(dstep <= 0){
        dstep = 1;          // using every sample
        dstepcounter = 0;
    }

    ringbuff     = lpf_get_ring_buffer(lpf_channel);
    ls.ringbuff  = ringbuff;
    ls.taps      = LPF_TAPS;
    ls.tapsmask  = LPF_TAPS_MASK;
    ls.lpf_fracbits = LPF_FRAC;

    p0 = lpf_p0[lpf_channel];
    top_pos = lpf_top[lpf_channel];

    dst = src;
    dstlen = 0;
    d0 = 0;
    while(d0 < length){
        sample = src[d0];
        dstepcounter += 1;
        if(dstepcounter >= dstep){
            dstepcounter = 0;
            ls.h = lpf_get_p_pointer(lpf_channel, top_pos & 0x03);
            ls.ringbuff_top = top_pos & ~0x03;
          #ifdef X86_USE_ASM_MMX
            dst[dstlen] = lpf_apply_calc_asm(&ls, (INT32)sample * p0);
          #else
            dst[dstlen] = lpf_apply_calc(&ls, (INT32)sample * p0);
          #endif
            dstlen += 1;
        }
        top_pos = (top_pos - 1) & LPF_TAPS_MASK;
        ringbuff[top_pos] = sample;
        d0 += 1;
    }
    lpf_top[lpf_channel] = top_pos;
    *downstepcounter = dstepcounter;

  #ifdef X86_USE_ASM_MMX
    do_emms();
  #endif

    return dstlen;
}


static int lpf_apply_sub_no_filter(int lpf_channel, INT16 *src, int length,
                                    int *downstep, int *downstepcounter)
{
//      return value is outlength
    INT16 *dst;
    int dstlen;
    int d0;
    int dstep;
    int dstepcounter;

    if(lpf_channel < 0) return 0;

    dstep        = *downstep;
    dstepcounter = *downstepcounter;
    if(dstep <= 1){
        dstep = 1;          // using every sample
        dstepcounter = 0;
    }

    dst = src;
    dstlen = 0;
    d0 = 0;
    while(d0 < length){
        dstepcounter += 1;
        if(dstepcounter >= dstep){
            dstepcounter = 0;
            dst[dstlen] = src[d0];
            dstlen += 1;
        }
        d0 += 1;
    }
    *downstepcounter = dstepcounter;

    return dstlen;
}


void lpf_apply(int lpf_channel, INT16* src, int length)
{
    int dstep  = 1;
    int dstepc = 0;

    if(use_filter){
        lpf_apply_sub(lpf_channel, src, length, &dstep, &dstepc);
    }else{
        lpf_apply_sub_no_filter(lpf_channel, src, length, &dstep, &dstepc);
    }
}


static int lpf_apply_downsampling(int lpf_channel, INT16* src, int length, SRCNV *s)
{
/* return value is outlength */
    int result;

    if(use_filter){
        result = lpf_apply_sub(lpf_channel, src, length, &s->down_step, &s->down_step_counter);
    }else{
        result = lpf_apply_sub_no_filter(lpf_channel, src, length, &s->down_step, &s->down_step_counter);
    }

    return result;
}



/********************************************************************
    upsampling(linear interpolation) & downsampling(LPF) section
********************************************************************/

int srcnv_init(int use_filter_flag)
{
    int d0;

    if(srcnv_init_flag == 0){
        srcnv_init_flag = 1;
        use_filter = use_filter_flag;       /* use_filter != 0, using lpf */
                                            /* use_filter == 0, no filter */
        lpf_current_used = 0;
        srcnv_current_used = 0;

        if(lpf_init_buffer() == 0){
            srcnv_init_flag = 0;
            return 0;
        }

        d0 = 0;
        while(d0 < SRCNV_MAX_CHANNELS){
            srcnv_data[d0].lpf_channel = -1;
            srcnv_data[d0].data_top = 0;
            d0 += 1;
        }
    }
    return 1;
}


void srcnv_exit(void)
{
    if(srcnv_init_flag != 0){
        srcnv_init_flag = 0;

        lpf_current_used = 0;
        srcnv_current_used = 0;
        lpf_exit_buffer();
    }
}


static int srcnv_change_sub(int srcnv_channel, double srcfreq, double dstfreq)
{
    SRCNV *s;
    int down_step;
    UINT64 d64_0;
    UINT64 d64_1;
    double lpf_freq;

    if(srcnv_channel < 0 || srcnv_current_used <= srcnv_channel) return -1;

#ifdef ERRORLOG
    logerror("SRCNV ch:%d, src:%d dst:%d\n", srcnv_channel, (int)srcfreq, (int)dstfreq);
#endif
    if(srcfreq < 1.0) srcfreq = 1.0;
    if(dstfreq < 1.0) dstfreq = 1.0;

    s = &srcnv_data[srcnv_channel];

    s->src_freq = srcfreq;
    s->dst_freq = dstfreq;
    if(srcfreq == dstfreq){
        s->no_op = 1;
#ifdef ERRORLOG
        logerror("SRCNV ch:%d, fs == fc, no-op\n", srcnv_channel);
#endif
    }else{
        s->no_op = 0;
        down_step = 1;
        while(down_step <= 65535){
            if( (srcfreq * 2.0) <= (dstfreq * down_step) ) break;
            down_step += 1;
        }
        if(down_step > 65535){
            s->no_op = 1;
        }else{
            d64_0 = (UINT64)dstfreq;
            d64_1 = (UINT64)(srcfreq * ((UINT64)1 << UPSAMPLING_FRAC));
            d64_0 = d64_1 / d64_0;
            if(d64_0 == 0) d64_0 = 1;
            s->src_dst_ratio = d64_0;
            d64_0 = (UINT64)(dstfreq * down_step);
            d64_0 = d64_1 / d64_0;
            if(d64_0 == 0) d64_0 = 1;
            s->upsampling_delta = (UINT32)d64_0;
            if(srcfreq > dstfreq){
                lpf_freq = dstfreq / 2;
            }else{
                lpf_freq = srcfreq / 2;
            }

            lpf_freq = lpf_freq - (lpf_freq * SRCNV_FC_ADJUST_RATIO);

            s->down_step = down_step;
            s->down_step_counter = 0;
            s->data[0] = 0;
            s->data[1] = 0;
            if(s->lpf_channel < 0){
                s->upsampling_current = 0;
                s->data_top = 0;
                s->lpf_channel = lpf_new(dstfreq * down_step, lpf_freq);
            }else{
                s->lpf_channel = lpf_change(s->lpf_channel, dstfreq * down_step, lpf_freq);
            }
        }
    }
    return srcnv_channel;
}


int srcnv_new(double srcfreq, double dstfreq)
{
    int srcnv_channel;

    if(srcnv_init_flag == 0) srcnv_init(1);     /* force to use lpf */

    if(srcnv_current_used >= SRCNV_MAX_CHANNELS) return -1;
    srcnv_channel = srcnv_current_used;
    srcnv_current_used += 1;

#ifdef ERRORLOG
    logerror("SRCNV_NEW ch:%d, used:%d\n", srcnv_channel, srcnv_current_used);
#endif
    srcnv_change_sub(srcnv_channel, srcfreq, dstfreq);

    return srcnv_channel;
}


int srcnv_change(int srcnv_channel, double srcfreq, double dstfreq)
{
#ifdef ERRORLOG
    logerror("SRCNV_CHANGE: ch:%d\n", srcnv_channel);
#endif
    srcnv_change_sub(srcnv_channel, srcfreq, dstfreq);

    return srcnv_channel;
}


static int srcnv_apply_upsampling(SRCNV *s)
{
    int samples;
    UINT32 d1;
    INT32 d2;
    UINT32 c;
    INT16  prev;
    INT16  *src;
    INT16  *dst;
    INT16  *data;
    int current_top_pos;
    int finished = 0;

    data = &s->data[0];
    dst = s->temp;
    src = s->src;
    if(src >= s->srcend){
        s->outbufferlen = 0;
        return 1;                               // finished
    }
    current_top_pos = s->data_top;              // current can be 0, 1
    c = s->upsampling_current;
    samples = 0;
    while(samples < s->tempbufferlen){
        data[current_top_pos] = *src;
        d1 = c >> (UPSAMPLING_FRAC - LINEAR_IP_FRACBITS);
        d1 = d1 & LINEAR_IP_FRACBITS_MASK;      // Q12 format
        prev = data[current_top_pos ^ 1];       // if current == 1 then prev == 0, vice versa
        d2 = data[current_top_pos] - prev;
        d2 = d2 * d1;
        d2 = d2 >> LINEAR_IP_FRACBITS;
        d2 = d2 + prev;
        dst[samples] = (INT16)d2;
        samples += 1;
        c  = c + s->upsampling_delta;           // delta must be smaller than (1 << UPSAMPLING_FRAC)
        d1 = c >> UPSAMPLING_FRAC;              // d1 = 0 or d1 = 1
        c  = c & UPSAMPLING_FRAC_MASK;
        current_top_pos = current_top_pos ^ d1;
        src += d1;
        if(src >= s->srcend){
            finished = 1;
            break;
        }
    }
    s->src = src;
    s->upsampling_outlen  = samples;
    s->upsampling_current = c;
    s->data_top = current_top_pos;

    return finished;
}


static int srcnv_apply_upsampling_no_interpolation(SRCNV *s)
{
    int samples;
    UINT32 d1;
    UINT32 c;
    INT16  *src;
    INT16  *dst;
    INT16  *data;
    int finished = 0;

    data = &s->data[0];
    dst = s->temp;
    src = s->src;
    if(src >= s->srcend){
        s->outbufferlen = 0;
        return 1;                               // finished
    }
    c = s->upsampling_current;
    samples = 0;
    while(samples < s->tempbufferlen){
        dst[samples] = *src;
        samples += 1;
        c  = c + s->upsampling_delta;           // delta must be smaller than (1 << UPSAMPLING_FRAC)
        d1 = c >> UPSAMPLING_FRAC;              // d1 = 0 or d1 = 1
        c  = c & UPSAMPLING_FRAC_MASK;
        src += d1;
        if(src >= s->srcend){
            finished = 1;
            break;
        }
    }
    s->src = src;
    s->upsampling_outlen  = samples;
    s->upsampling_current = c;

    return finished;
}



int srcnv_apply_init(int srcnv_channel, INT16* src, int length)
{
    SRCNV *s;

    if(srcnv_channel < 0 || srcnv_current_used <= srcnv_channel){
#ifdef ERRORLOG
        logerror("srcnv_apply_init ch error\n");
#endif
        return 0;
    }

    s = &srcnv_data[srcnv_channel];

    s->src    = src;
    s->srcend = src + length;
    s->temp              = &srcnv_buffer[0];
    s->tempbufferlen     = SRCNV_BUFFER_LENGTH;
    s->upsampling_outlen = 0;

    s->outbuffer    = src;          /*  for no-op (Fs == Fc) */
    s->outbufferlen = length;       /*  src_data == dst_data */

    s->finished = 0;

    return 1;
}


int srcnv_apply_exit(int srcnv_channel)
{
    SRCNV *s;

    if(srcnv_channel < 0 || srcnv_current_used <= srcnv_channel){
#ifdef ERRORLOG
        logerror("srcnv_apply_init ch error\n");
#endif
        return 0;
    }

    s = &srcnv_data[srcnv_channel];
    s->src    = 0;
    s->outbuffer    = &srcnv_buffer[0];
    s->outbufferlen = 0;

    s->finished = 1;

    return 1;
}


int srcnv_apply(int srcnv_channel)
{
    SRCNV *s;

    if(srcnv_channel < 0 || srcnv_current_used <= srcnv_channel){
#ifdef ERRORLOG
        logerror("srcnv_apply ch error\n");
#endif
        return 1;                       // finished
    }
    s = &srcnv_data[srcnv_channel];
    s->finished = 1;
    if(s->no_op != 0){
        return s->finished;             // finished
    }
    if(s->src == 0 || s->src >= s->srcend){
        return s->finished;             // finished
    }

    if(use_filter){
        s->finished = srcnv_apply_upsampling(s);
    }else{
        s->finished = srcnv_apply_upsampling_no_interpolation(s);
    }
    s->outbufferlen  = lpf_apply_downsampling(s->lpf_channel, s->temp,
                                s->upsampling_outlen, s);
    s->outbuffer = &srcnv_buffer[0];

    return s->finished;
}


int srcnv_is_finished(int srcnv_channel)
{
    if(srcnv_channel < 0 || srcnv_current_used <= srcnv_channel){
#ifdef ERRORLOG
        logerror("srcnv_apply ch error\n");
#endif
        return 1;                       // finished;
    }

    return srcnv_data[srcnv_channel].finished;
}


INT16* srcnv_get_outbuffer_address(int srcnv_channel)
{
    if(srcnv_channel < 0 || srcnv_current_used <= srcnv_channel){
#ifdef ERRORLOG
        logerror("srcnv_apply ch error\n");
#endif
        return 0;
    }
    return srcnv_data[srcnv_channel].outbuffer;
}


int srcnv_get_outbuffer_length(int srcnv_channel)
{
    if(srcnv_channel < 0 || srcnv_current_used <= srcnv_channel){
#ifdef ERRORLOG
        logerror("srcnv_apply ch error\n");
#endif
        return 0;
    }
    return srcnv_data[srcnv_channel].outbufferlen;
}

int srcnv_get_filter_taps(void)
{
    return LPF_TAPS;
}



int srcnv_get_src_sample_length(double srcfreq, double dstfreq, int dst_length)
{
    UINT64 d64_0;
    UINT64 d64_1;

    d64_0 = (UINT64)dstfreq;
    d64_1 = (UINT64)(srcfreq * ((UINT64)1 << UPSAMPLING_FRAC));
    d64_0 = d64_1 / d64_0;

    d64_0 = d64_0 * (UINT64)dst_length;
    d64_0 = d64_0 >> UPSAMPLING_FRAC;

    return (int)d64_0;
}
